
#define ADC_MAX 1024
#define SUM_MAX 500
#define PAUSE_CNT 500

static uint32_t   o_last_milli=0;
static float      vals[ADC_MAX+1];
static int        o_count=0;
static double     o_Wmin=0;
static int        o_pause=0;

static int minValue=0, maxValue=0;
static uint32_t sum_adc=0;
static uint32_t lsum_adc=0;
int xset=1;
static float   adc_fac=1.0917;  // 1.0917

extern time_t getCurrentSecs( void );

static uint32_t tieflast=0;
static float    adc_tief=0;

void do_measurement( void )
{
  float     adc_read;
  uint32_t  now;
  float     adc_avg;
  uint32_t  secs;
  uint32_t  diff;

  if ( o_pause )
  {
    o_pause--;
    return;
  }

  now = millis() | 0x80000000;

  if ( !tieflast )
  {
    tieflast = now;
    adc_tief = analogRead(A0);
  }

  if ( now < tieflast )
  {
    tieflast &= 0x7fffffff;    /* remove bit31 */
    tieflast |= 1;             /* make != 0 */
  }

  diff=now-tieflast;
  tieflast = now | 0x80000000;

  adc_read = analogRead(A0);
  if ( adc_read > ADC_MAX )
    return;

  adc_tief = (adc_read - adc_tief)*(diff)/10000+adc_tief;
  tieflast = now;

  if ( xset && ( o_count < 10 ))
    vals[ o_count ] = adc_read;

  adc_read *= adc_fac;           // 469 -> 512

  if ( adc_read > ADC_MAX )
    return;
  if ( adc_read < 512 )
    adc_read = 512 - adc_read;
  else
    adc_read -= 512;

  if ( !o_count )
  {
    minValue = adc_read;
    maxValue = adc_read;
    sum_adc = adc_read;
  }
  else
  {
    sum_adc += adc_read;
    if ( adc_read < minValue )
      minValue = adc_read;
    if ( adc_read > maxValue )
      maxValue = adc_read;
  }

  o_count++;

  if ( o_count < SUM_MAX )
    return;

  lsum_adc=sum_adc;
  adc_avg = sum_adc / o_count;

  float mVperAmp=100;  // 20A=100, 5A=185, 30A=66
  float Ieff;
  int   ACSoffset=2500;
  //float voltage = (maxValue - minValue)* 5000 * 1.094 / 1024.0;
  float voltage = 4.883 * adc_avg;   // scale 512 -> 2500
  Ieff = voltage/mVperAmp;
    
  //float Ieff = (voltage - ACSoffset) / mVperAmp;
  float Ueff = voltage;

  float P = Ieff*230;

  o_count=0;
  xset=0;
  now = millis() | 0x80000000;
  o_last_milli |= 0x80000000;
  if ( now < o_last_milli )
  {
    o_last_milli &= 0x7fffffff;    /* remove bit31 */
    o_last_milli |= 1;             /* make != 0 */
  }
  o_Wmin += P*(now-o_last_milli)/60000.0;  /* P * anz_milli / 60000 : WattMinute */
  o_last_milli = now;
 
  // save
  secs=getCurrentSecs();
  if ( secs )
	memAddValue( o_Wmin, secs );
  // publish
  double ed=o_Wmin / 60000; /* khW */
  _iobSetFloatVar( IDX_WM, o_Wmin, SRC_BTN );
  _iobSetFloatVar( IDX_KWH, ed, SRC_BTN );
  _iobSetFloatVar( IDX_POWER, (double)P, SRC_BTN );
  _iobSetFloatVar( IDX_VOLTAGE, (double)Ueff, SRC_BTN );
  _iobSetFloatVar( IDX_CURRENT, (double)Ieff, SRC_BTN );
#ifdef SERIAL_DEBUG
Serial.printf("Wmin    : %d\n",(int)o_Wmin);
Serial.printf("KWh     : %d\n",(int)ed);
Serial.printf("Power   : %d\n",(int)P);
Serial.printf("Current : %d\n",(int)Ieff);
Serial.printf("Voltage : %d\n",(int)Ueff);
#endif
  o_pause = PAUSE_CNT;
}

void acs712_resetWMin( int val )
{
	o_Wmin=val;
}

void acs712_adc_Init( void )
{
  o_last_milli = millis() | 0x80000000;
  o_count=0;
  o_Wmin = 0;

  do_measurement( );
}

void acs712_adc_loop( void )
{
  uint32_t  now=millis() | 0x80000000;
  uint32_t  diff;

  do_measurement( );
}

String acs712_debug( int full )
{
  String res=F("<BR><table cellspacing=2>");
  int   sum=0;
  int   i;
  if ( xset )
    return (String)"";
  for(i=0;i<10;i++)
  {
    sum+=vals[i];
    if ( full )
    {
      res+=F("<tr><td>");
      res+=i;
      res+=F("</td><td>");
      res+=vals[i];
      res+=F("</td><td>");
      res+=adc_fac * vals[i];
      res+=F("</td></tr>");
    }
  }
  res+=F("<tr><td>acs-sum</td><td>");
  res+=sum;
  res+=F("</td></tr><tr><td>acs-tief</td><td>");
  res+=(String)(adc_tief);
  res+=F("</td></tr></table>");
  xset=1;
  return res;
}
