
#include <ESP8266WiFi.h>
#include <EEPROM.h>
#include <ESP8266httpUpdate.h>

extern "C" {
#include "user_interface.h"
#include "iob.h"
#include "config.h"
}

class WiFiQueue
{
  public:
    WiFiClient        client;
    int               fd;       /* read from file */
    int               step;
    int               wrong_cnt;
    int               used;
    int               sum;
    int               s_done;
    int               len;
    void              *data;
};

class Config
{
  public:
    String    DevName;
    String    WLan;
    String    Password;
    byte      is_dhcp;
    byte      use_iob;
    byte      use_remiob;
    String    IP;
    String    Mask;
    String    Gateway;
    String    DNS;
    String    wlan_list;
    String    cur_wlan;
    String    iobIP;
    uint32_t  adcnull;
    byte      as_ap;	/* set if WLan == "local.AP" */
};

#define CONF_DONE   112
#define CONF_RECONF 113

#define SEND_PCK_SIZE 512

extern int	__t_open( char *fname );
extern unsigned long	__t_getsize( int fd );
extern int	__t_read( int fd, void *to, int sz );
extern int	__t_close( int fd );

#define M_ESP_SETUP 0
#define M_ESP_RUN   1

extern int runmode;   /* configure esp */

WiFiServer server(80);

static WiFiQueue   clist[10];
Config      config;

static void readRTC( void )
{
#if 0
  while(system_rtc_mem_read(RTC_BASE,rtcmem,RTC_RAM_ARRAY_SIZE)!=true)
        delay(1);
#endif
}

static void writeRTC( void )
{
#if 0
  rtcmem[4] = RTC_VALID;
  memcpy(rtcmem,&hlw.kWhtoday,sizeof(long));
  while(system_rtc_mem_write(RTC_BASE,rtcmem,RTC_RAM_ARRAY_SIZE)!=true)
        delay(1);
  yield();
  delay(5);
#endif
}

static void prepareWiFiSetup( int noap )
{
#ifdef SERIAL_DEBUG
  Serial.println(F("Scan networks"));
#endif
  int   percentage;
  int   chk = 0, k, kp;
  int   i, n=0;

  n = WiFi.scanNetworks(false,true);
#ifdef SERIAL_DEBUG
  Serial.println(F("Scan networks ... done"));
#endif

  if ( !noap )
    WiFi.mode(WIFI_STA);

#ifdef SERIAL_DEBUG
  Serial.print(F("set AP done, ssid-cnt="));
  Serial.println(n);
#endif

  if ( !noap && !n )
  {
    writeRTC();
    ESP.restart();
    delay(10);
  }

  config.wlan_list = "";
  config.cur_wlan = "";
  for ( i = 0; i < n; i++ )
  {
    // WiFi.SSID(i);
    // WiFi.RSSI(i);
    // WiFi.encryptionType(i) == ENC_TYPE_NONE ?
    percentage = (WiFi.RSSI(i) + 100) * 2;
    if ( percentage < 0 )
      percentage = 0;
    else if ( percentage > 100 )
      percentage = 100;
    kp = i;
    /* check if ssid already reported */
    for ( k = 0; k < i; k++ )
    {
      if ( WiFi.SSID(i).equals(WiFi.SSID(k)) )
        break;
    }
    if ( k != i )
      continue;
    /* check if ssid is reported again */
    for ( k = i + 1; k < n; k++ )
    {
      if ( WiFi.SSID(i).equals(WiFi.SSID(k)) )
      {
        int  kperc = (WiFi.RSSI(k) + 100) * 2;
        if ( kperc < 0 )
          kperc = 0;
        else if ( kperc > 100 )
          kperc = 100;
        if ( kperc > percentage )
        {
          kp = k;
          percentage = kperc;
        }
      }
    }

    config.wlan_list += "<li id=\"";
    config.wlan_list += WiFi.SSID(i);
    config.wlan_list += "\" data-strength=\"";
    config.wlan_list += percentage;
    config.wlan_list += "\"";
    if ( !chk && config.WLan.equals(WiFi.SSID(i)) )
    {
      config.wlan_list += " class=\"selected\"";
      config.cur_wlan = WiFi.SSID(i);
      chk = 1;
    }
    config.wlan_list += "></li>\n";
#ifdef SERIAL_DEBUG
    Serial.print( F("[") );
    Serial.print( kp );
    Serial.print( F("] : ") );
    Serial.print( WiFi.SSID(i) );
    Serial.print( F(" (" ) );
    Serial.print( WiFi.RSSI(kp) );
    Serial.print( F(" = " ) );
    Serial.print( percentage );
    Serial.print( F("% ) ") );
    Serial.println( (WiFi.encryptionType(i) == ENC_TYPE_NONE) ? " " : "*" );
#endif
  }
/* now add local-AP as option */
  config.wlan_list += "<li id=\"";
  config.wlan_list += "local.AP";
  config.wlan_list += "\" data-strength=\"100\"";
  if ( !chk && config.WLan.equals("local.AP") )
  {
    config.wlan_list += " class=\"selected\"";
    config.cur_wlan = "local.AP";
  }
  config.wlan_list += "></li>\n";

  if ( noap )
    return;

  WiFi.softAP( SETUP_SSID, SETUP_PASSWD );
  delay(100);
#ifdef SERIAL_DEBUG
  Serial.println(F("SoftAP started") );
#endif
}

static void prepareWiFiConfSetup( void )
{
  server.begin();
}

static int SaveParam( String Param )
{
  int   rc = 0;
  int addr = Param.indexOf('=');

  if ( addr == -1 )
    return rc;
  String Name = Param.substring(0, addr);
  String Value = Param.substring(addr + 1);

  Value.replace("%3A", ":");
  Value.replace("%20", " ");
  Value.replace("%21", "!");
  Value.replace("%2B", "+");
  Value.replace("%2C", ",");
  if ( Name.startsWith("times") )
    Value.replace("+", " ");
#ifdef SERIAL_DEBUG
  Serial.printf("SaveParam: '%s'='%s'\n", Name.c_str(), Value.c_str());
#endif

  if ( Name.equals("devicename") && Value.length() && (rc = 1) )
    config.DevName = Value;
  else if ( Name.equals("wlan") && Value.length() && (rc = 2) )
    config.WLan = Value;
  else if ( Name.equals("password") && Value.length() && (rc = 4) )
    config.Password = Value;
  else if ( Name.equals("lantype") && Value.length() )
  {
    config.is_dhcp = Value.equals("dhcp") ? 1 : 0;
    rc = config.is_dhcp ? 8 : 0;
  }
  else if ( Name.equals("ip") && Value.length() && (rc = 16) )
    config.IP = Value;
  else if ( Name.equals("mask") && Value.length() && (rc = 32) )
    config.Mask = Value;
  else if ( Name.equals("gateway") && Value.length() && (rc = 64) )
    config.Gateway = Value;
  else if ( Name.equals("dns") && Value.length() && (rc = 128) )
    config.DNS = Value;
  else if ( Name.equals("iob-server") )
  {
    config.use_iob = Value.equals("on") ? 1 : 0;
    rc = 256;
  }
  else if ( Name.equals("iob-client") )
  {
    config.use_remiob = Value.equals("on") ? 1 : 0;
    rc = 512;
  }
  else if ( Name.equals("iobip") && Value.length() && (rc = 2048) )
    config.iobIP = Value;
  else if ( Name.equals("adcnull") && Value.length() )
    config.adcnull = (uint32_t)Value.toInt();

  return rc;
}

static int setup_done = 0;

static void EESaveString( int pos, String val, int maxlen )
{
  int i;

  for ( i = 0; i < val.length() && i < maxlen; i++ )
    EEPROM.write(pos + i, val[i]);
  if ( i < maxlen )
    EEPROM.write(pos + i, 0); /* end of string */
}

static void EESaveInt32( int pos, uint32_t val )
{
  int i;
  for ( i = 0; i < 4; i++ )
  {
    EEPROM.write(pos + i, val & 0xff);
    val >>= 8;
  }
}

static void EESaveInt16( int pos, uint32_t val )
{
  int i;
  for ( i = 0; i < 2; i++ )
  {
    EEPROM.write(pos + i, val & 0xff);
    val >>= 8;
  }
}

static void EESaveIP( int pos, String val, int incl_port )
{
  byte  ip[4];
  int   port = 0;
  int   i, a = 0, b;
  int   pPos = -1;

  pPos = val.indexOf(':');
  if ( pPos == -1 )
  {
    pPos = val.indexOf("%3A");
    if ( pPos != -1 )
      pPos += 3;
  }
  if ( pPos != -1 )
    port = val.substring(pPos).toInt();

  for ( i = 0; i < 3; i++ )
  {
    b = val.indexOf('.', a);
    ip[i] = val.substring(a, b).toInt();
    a = b + 1;
  }
  if ( pPos != -1 )
    ip[i] = val.substring(a, pPos).toInt();
  else
    ip[i] = val.substring(a).toInt();

  for ( i = 0; i < 4; i++)
    EEPROM.write(pos + i, ip[i]);

  if ( incl_port )
    EESaveInt16(pos + 4, port);
}

#define EE_CONFIG   0   /* 1 byte */
#define EE_DHCP     1   /* 1 byte */
#define EE_DEVNAME  2   /* 24 byte */
#define EE_WLAN     26  /* 32 byte */
#define EE_PASSWD   48  /* 64 byte */
#define EE_IP       112 /* 4 byte */
#define EE_MASK     116 /* 4 byte */
#define EE_GATEWAY  120 /* 4 byte */
#define EE_DNS      124 /* 4 byte */
#define EE_USEIOB   128 /* 1 byte */
#define EE_USERIOB  129 /* 1 byte */
#define EE_IOBIP    130 /* 4 byte */
#define EE_ADCNULL  134 /* 4 byte */
#define EE_FREE     138
#define EE_MEMORIZE 512 /* 285 bytes */

static void Save2EEProm( int timer_only )
{
  EEPROM.begin(512);
  delay(10);
  EEPROM.write(EE_DHCP, config.is_dhcp);
  EESaveString(EE_DEVNAME, config.DevName, 24);
  EESaveString(EE_WLAN, config.WLan, 32);
  EESaveString(EE_PASSWD, config.Password, 64);
  EESaveIP(EE_IP, config.IP, 0);
  EESaveIP(EE_MASK, config.Mask, 0);
  EESaveIP(EE_GATEWAY, config.Gateway, 0);
  EESaveIP(EE_DNS, config.DNS, 0);
  EEPROM.write(EE_USEIOB, config.use_iob);
  EEPROM.write(EE_USERIOB, config.use_remiob);
  EESaveIP(EE_IOBIP, config.iobIP, 0);
  EESaveInt32(EE_ADCNULL, config.adcnull);
 
  EEPROM.write(EE_CONFIG, CONF_DONE);
  EEPROM.commit();
  EEPROM.end();
  delay(10);
}

String EEReadString(int pos, int maxlen )
{
  String Value = "";
  char   z;
  int i;
  for ( i = 0; i < maxlen; i++ )
  {
    z = char(EEPROM.read(pos + i));
    if ( !z )
      break;
    Value += z;
  }
  return Value;
}

int EEReadInt16( int pos )
{
  int val;
  val = EEPROM.read(pos + 1) << 8 + EEPROM.read(pos);
  return val;
}

int EEReadInt32( int pos )
{
  int val=0;
  int i;
  pos += 3;

  for( i=0; i<4; i++ )
  {
    val <<= 8;
    val |= EEPROM.read(pos);
    pos--;
  }
  return val;
}

String EEReadIP( int pos, int incl_port )
{
  byte      ip[4];
  int       i;
  uint32_t  port = 0;
  String    val;

  for ( i = 0; i < 4; i++ )
    ip[i] = EEPROM.read(pos + i);
  if ( incl_port)
    port = EEReadInt16(pos + 4);
  val += ip[0];
  val += ".";
  val += ip[1];
  val += ".";
  val += ip[2];
  val += ".";
  val += ip[3];
  if ( port )
  {
    val += ":";
    val += port;
  }
  return val;
}

static void ReadEEProm( void )
{
  byte done;

  memset(&config, 0, sizeof(config));

  EEPROM.begin(512);
  EEPROM.get(EE_CONFIG, done);
  if (( done != CONF_DONE ) &&
      ( done != CONF_RECONF ))
  {
    config.is_dhcp = 1;

    EEPROM.end();
    return;
  }

  EEPROM.get(EE_DHCP, config.is_dhcp);
  config.DevName = EEReadString(EE_DEVNAME, 24);
  config.WLan = EEReadString(EE_WLAN, 32);
  config.Password = EEReadString(EE_PASSWD, 64);
  config.IP = EEReadIP(EE_IP, 0);
  config.Mask = EEReadIP(EE_MASK, 0);
  config.Gateway = EEReadIP(EE_GATEWAY, 0);
  config.DNS = EEReadIP(EE_DNS, 0);
  EEPROM.get(EE_USEIOB, config.use_iob);
  EEPROM.get(EE_USERIOB, config.use_remiob);
  config.iobIP = EEReadIP(EE_IOBIP, 0);
  config.adcnull = EEReadInt32(EE_ADCNULL);
  if (( config.adcnull > 4000 ) && ( config.adcnull < 6000 ))
    adc_fac = 5120.0 / config.adcnull;
  EEPROM.end();
#ifdef SERIAL_DEBUG
  Serial.println("eeprom-config:");
  Serial.printf("DevName='%s'\n", config.DevName.c_str());
  Serial.printf("WLan='%s', Password='%s'\n", config.WLan.c_str(), config.Password.c_str());
#endif

  if ( done == CONF_DONE )
    runmode = M_ESP_RUN;
}

static void ConfigReplace( String *text )
{
  text->replace("${DEVNAME}", config.DevName);

  text->replace("${WLAN}", config.wlan_list);
  text->replace("${PASSWD}", config.Password);
  text->replace("${DHCP}", config.is_dhcp ? "checked=\"checked\"" : "");
  text->replace("${STATIC}", config.is_dhcp ? "" : "checked=\"checked\"");
  text->replace("${IPADDR}", config.IP);
  text->replace("${MASK}", config.Mask);
  text->replace("${GATEWAY}", config.Gateway);
  text->replace("${DNS}", config.DNS);
  text->replace("${IPHIDDEN}", config.is_dhcp ? " class=\"hidden\"" : "");
  text->replace("${CURWLAN}", config.cur_wlan);

  text->replace("${USEIOB}", config.use_iob ? "checked=\"checked\"" : "");
  text->replace("${USEREMI}", config.use_remiob ? "checked=\"checked\"" : "");
  text->replace("${IOBCLHIDDEN}", config.use_remiob ? "" : " class=\"hidden\"");
  text->replace("${IOBIP}", config.iobIP);
  text->replace("${ADCNULL}", (String)config.adcnull);
}

static void SendbackHdrOnly( int i )
{
     String HtmlHeader;

      HtmlHeader=F("HTTP/1.1 200 OK\r\nServer: esp8266\r\nConnection: close\r\n");
      HtmlHeader+=F("Access-Control-Allow-Origin: *\r\n");
      HtmlHeader+=F("Content-Length: 0\r\nContent-Type: text/html\r\n\r\n");

      clist[i].client.print(HtmlHeader);
      clist[i].step=99;
}

static void SendbackVal( int i, String val )
{
     String HtmlHeader;

      HtmlHeader=F("HTTP/1.1 200 OK\r\nServer: esp8266\r\nConnection: close\r\n");
      HtmlHeader+=F("Access-Control-Allow-Origin: *\r\n");
      HtmlHeader+=F("Content-Length: ");
      HtmlHeader+=val.length();
      HtmlHeader+=F("\r\nContent-Type: text/html\r\n\r\n");

      HtmlHeader+=val;

      clist[i].client.print(HtmlHeader);
      clist[i].step=99;
}

static byte   w_connected = 0;
static byte   w_trys = 0;
static uint16_t w_connloops = 0;
int mxmode=0;

void DoQueuedWiFiClient( int i )
{
  char setup_requested=0;
  char info_requested=0;
  int full;

  if ( !clist[i].client )
  {
    clist[i].step = 99;
    return;
  }
  if ( clist[i].step == 0 )
  {
    if (clist[i].client.connected())
      clist[i].step = 1;
    else
    {
      clist[i].wrong_cnt++;
      if ( clist[i].wrong_cnt > 100 )
      {
        clist[i].step = 99;
#ifdef SERIAL_DEBUG
        Serial.println("wrong count reached");
#endif
        return;
      }
    }
  }

  if ( !clist[i].client.connected() )
  {
    clist[i].step = 99;
    return;
  }
  if ( clist[i].step == 1 )
  {
    if ( !clist[i].client.available() )
      return;

    String RequestString = clist[i].client.readStringUntil('\n');
    int addr_end = RequestString.indexOf(' ', 5); // POST x......, GET x......
    if ( addr_end == -1 )
    {
      clist[i].step = 99;
      return;
    }

    /* read until \r\n\r\n */
    int max = 100;
    while ( max-- )
    {
      String Dummy = clist[i].client.readStringUntil('\n');
      if ( Dummy.length() < 2 )
        break;

      delay(5);
    }

    RequestString = RequestString.substring(0, addr_end);
#ifdef SERIAL_DEBUG
    Serial.print(F("["));
    Serial.print(i);
    Serial.print(F("] : "));
    Serial.println( RequestString );
#endif
    if ( RequestString.startsWith("POST /index.html") )
    {
      // read parameters
      String ParamList = clist[i].client.readStringUntil('\n');
      String Param;
      int    addr_start = 0;
      int    addr_end;
      int    reached = 0;

      /* unset values will not transferred */
      config.is_dhcp = 0;
      config.use_iob = 0;
      config.use_remiob = 0;
      config.adcnull = 469;

      while ( 1 )
      {
        addr_end = ParamList.indexOf('&', addr_start);
        if ( addr_end != -1 )
          Param = ParamList.substring(addr_start, addr_end);
        else
          Param = ParamList.substring(addr_start);
        reached |= SaveParam( Param );
        if ( addr_end == -1 )
          break;
        addr_start = addr_end + 1;
      }
#ifdef SERIAL_DEBUG
      Serial.printf("reached=%04x\n", reached);
#endif
      /***
         okay : 0 0000 1111 0111		 last 8 bits
         okay : 0 0000 0000 1111		 last 8 bits
         okay : 0 1010 0000 0000		 first 5 bits
         okay : 1 0100 0000 0000		 first 5 bits
      ***/
      setup_done = 0;
      setup_done |= ((reached & 0x0f) == 0x0f) ? 1 : 0;
      setup_done |= ((reached & 0xf7) == 0xf7) ? 1 : 0;
      while ( setup_done )
      {
        if (( reached & 0x0c00 ) == 0x0200 )
        {
          setup_done = 0;
          break;
        }
        if (( reached & 0x1400 ) == 0x0400 )
          setup_done = 0;
        break;
      }
      RequestString = setup_done ? "GET /ready.html" : "GET /index.html";
      Save2EEProm(0);
    }

    if ( RequestString.startsWith("POST /timer.html") )
    {
      // read parameters
      String ParamList = clist[i].client.readStringUntil('\n');
      String Param;
      int    addr_start = 0;
      int    addr_end;
      int    reached = 0;

      while ( 1 )
      {
        addr_end = ParamList.indexOf('&', addr_start);
        if ( addr_end != -1 )
          Param = ParamList.substring(addr_start, addr_end);
        else
          Param = ParamList.substring(addr_start);
        reached |= SaveParam( Param );
        if ( addr_end == -1 )
          break;
        addr_start = addr_end + 1;
      }
#ifdef SERIAL_DEBUG
      Serial.printf("reached=%04x\n", reached);
#endif
      setup_done = 0;
      Save2EEProm(1);
      RequestString = "GET /ready.html";
    }
    full=0;

    if ( RequestString.equals("GET /index.html") ||
         RequestString.equals("GET /") )
    {
      if ( RequestString.indexOf("?full") != -1 )
        full=1;
      if ( runmode == M_ESP_RUN )
        info_requested=1;
      else
        setup_requested=1;
    }
    else if ( RequestString.equals("GET /setup.html") )
    {
       setup_requested=1;
    }
    else if ( RequestString.startsWith("GET /info.html") )
    {
      info_requested=1;
      if ( RequestString.indexOf("?full") != -1 )
        full=1;
    }
    if ( setup_requested )
    {
      int fd = __t_open("index.html");
      String HtmlSite, HtmlHeader;

      if ( !config.wlan_list.length() )
        prepareWiFiSetup(1);

      if ( fd == -1 )
      {
        SendbackHdrOnly(i);
#ifdef SERIAL_DEBUG
        Serial.println("**** file not found 'index.html'");
#endif
        return;
      }

      char buff[520];
      int xb = __t_read(fd, buff, 512);

      while ( xb > 0 )
      {
        buff[xb] = 0;
        HtmlSite += buff;
        xb = __t_read(fd, buff, 512);
      }
      __t_close(fd);

      ConfigReplace(&HtmlSite);
#ifdef SERIAL_DEBUG
      Serial.println(HtmlSite);
#endif

      HtmlHeader = F("HTTP/1.1 200 OK\r\n");
      HtmlHeader += F("Content-Length: ");
      HtmlHeader += HtmlSite.length();
      HtmlHeader += F("\r\n");
      HtmlHeader += F("Content-Type: text/html\r\n");
      HtmlHeader += F("Access-Control-Allow-Origin: *\r\n");
      HtmlHeader += F("Connection: close\r\n");
      //        HtmlHeader += F("Connection: keep-alive\r\n");
      //        HtmlHeader += F("Keep-Alive: timeout=10, max=2000\r\n");
      HtmlHeader += F("\r\n");

      clist[i].client.print(HtmlHeader);
      {
        const char *p = HtmlSite.c_str();
        int  len = HtmlSite.length();
        int  sz;
 
        sz = len;
        if ( sz > 512 )
          sz = 512;
        clist[i].client.write(p, sz);
#ifdef SERIAL_DEBUG
      Serial.print(sz);
      Serial.println(" bytes written");
#endif
        p += sz;
        len -= sz;
        clist[i].data=malloc(len);
        memcpy(clist[i].data,p,len);
        clist[i].step = 3;
        clist[i].s_done=0;  /* already sent */
        clist[i].len=len;
      }

      return;
    }

    if ( info_requested )
    {
      String HtmlHeader;
      String HtmlBody;
      String HtmlHead;
      char   data[64];
      int	fd = __t_open( "head/chart" );
      int	sz = __t_getsize(fd);
      char  buff[520];

      HtmlHead=F("<html><head><meta charset=\"utf-8\"/><title>");
			HtmlHead+=config.DevName;
			HtmlHead+=F("</title>");

      HtmlBody=F("<body onload=\"graph( 'per hour', 'd24h', 'rgb(255,50,110', 1,");
      HtmlBody+=Mem24Values();
/* insert values */
      HtmlBody+=F("); graph( 'last-30-days', 'd30d', 'rgb(210,52,255', 1, 29,");
      HtmlBody+=MemDayValues();
      HtmlBody+=F("); graph( 'per month', 'd12m', 'rgb(110,230,230', 0,");
      HtmlBody+=MemMonthValues();
      HtmlBody+=F("); curve( 'watt-per-minute', 'd60m', 'rgb(230,180,100', 1,");
      HtmlBody+=MemMinuteValues();
      HtmlBody+=F("); \">");
	    HtmlBody+=F("IP Address: ");
      HtmlBody+=WiFi.localIP().toString();
      HtmlBody+=F("<BR>FlashSizebyChipID: ");
      HtmlBody+=ESP.getFlashChipSizeByChipId();
      HtmlBody+=F("<BR>Device name: ");
      HtmlBody+=config.DevName;
      HtmlBody+=F("<BR>Modul typ: ");
      HtmlBody+=MODUL;
      HtmlBody+=F("<BR>Version: ");

      strcpy(data,VERSION_STRG);
      HtmlBody+=(String)data;
      
      HtmlBody+=F("<BR>boot-reason: ");
      HtmlBody+=boot_reason_text();
      HtmlBody+=F("<BR>life-millis: ");
      HtmlBody+=millis();
      HtmlBody+=F("<BR>ADC-null   : ");
      HtmlBody+=config.adcnull;
      HtmlBody+=F("<BR>ADC-fac    : ");
      HtmlBody+=adc_fac*100;
      HtmlBody+=F("<BR>disco_cnt    : ");
      HtmlBody+=disco_count;

      HtmlBody+=iobDebug( full );
//      HtmlBody+=memDebug();
      HtmlBody+=acs712_debug( full );
      HtmlBody+=timerDebug( full );

      HtmlBody+=F("<BR><canvas id=\"d60m\" width=\"800\" height=\"420\"></canvas>");
      HtmlBody+=F("<BR><canvas id=\"d24h\" width=\"800\" height=\"250\"></canvas>");
      HtmlBody+=F("<BR><canvas id=\"d30d\" width=\"800\" height=\"250\"></canvas>");
      HtmlBody+=F("<BR><canvas id=\"d12m\" width=\"800\" height=\"240\"></canvas>");
      HtmlBody+=F("</body></html>");

      HtmlHeader=F("HTTP/1.1 200 OK\r\nServer: esp8266\r\nConnection: close\r\n");
      HtmlHeader+=F("Access-Control-Allow-Origin: *\r\n");
      HtmlHeader+=F("Content-Length: ");
      HtmlHeader+=(String)(HtmlHead.length()+sz+HtmlBody.length());
      HtmlHeader+=F("\r\nContent-Type: text/html\r\n\r\n");

      clist[i].client.print(HtmlHeader);
      clist[i].client.print(HtmlHead);
	  int xsz=sz;
	  while( xsz )
	  {
        int xb = __t_read(fd, buff, SEND_PCK_SIZE);
		if ( xb > 0 )
		{
		  clist[i].client.write((const uint8_t*)buff, xb);
		  xsz-=xb;
		}
		else
			break;
	  }
      __t_close(fd);
      clist[i].client.print(HtmlBody);
      clist[i].step=99;
      return;
    }

    if ( RequestString.startsWith("GET /UPDATE?") )
    {

      SendbackHdrOnly(i);
#ifdef SERIAL_DEBUG
      Serial.printf("**** file '%s'\n",(char*) RequestString.substring(12).c_str());
#endif
      _doUpdate(RequestString.substring(12).c_str());
      return;
    }

    if ( RequestString.startsWith("GET /ZERO") )
    {

      SendbackHdrOnly(i);
      memorizeReset();
      return;
    }

    if ( RequestString.startsWith("GET /REBOOT") || RequestString.startsWith("GET /reboot"))
    {
      SendbackHdrOnly(i);
#ifdef SERIAL_DEBUG
      Serial.printf("REBOOT initiated\n");
#endif
      writeRTC();
			memorizeSave();
      ESP.reset();
      return;
    }

    if ( RequestString.startsWith("GET /") )
    {
      int punkt = RequestString.lastIndexOf('.');
      int fd = __t_open( (char*) RequestString.substring(5).c_str());
      String ext = RequestString.substring(punkt + 1);
      String HtmlHeader, HtmlSite;

      if ( ext == "js" )
        ext = F("javascript");

      if ( fd == -1 )
      {
        HtmlHeader = F("HTTP/1.1 404 Not Found\r\nServer: esp8266\r\nConnection: close\r\n");
        HtmlHeader += F("Access-Control-Allow-Origin: *\r\n");
        HtmlHeader += F("Content-Length: 0\r\nContent-Type: text/html\r\n\r\n");
#ifdef SERIAL_DEBUG
        Serial.printf("**** file not found '%s'\n", (char*) RequestString.substring(5).c_str());
#endif

        clist[i].client.print(HtmlHeader);
        clist[i].step = 99;
        return;
      }

      char buff[SEND_PCK_SIZE];
      int sz = __t_getsize(fd);

      HtmlHeader = F("HTTP/1.1 200 OK\r\n");
      HtmlHeader += F("Content-Length: ");
      HtmlHeader += sz;
      HtmlHeader += F("\r\n");
      HtmlHeader += F("Content-Type: text/");
      HtmlHeader += ext;
      HtmlHeader += F("\r\n");
      HtmlHeader += F("Access-Control-Allow-Origin: *\r\n");
      HtmlHeader += F("Connection: close\r\n");
      //        HtmlHeader += F("Connection: keep-alive\r\n");
      //        HtmlHeader += F("Keep-Alive: timeout=10, max=2000\r\n");


      HtmlHeader += F("\r\n");

      clist[i].client.print(HtmlHeader);

      clist[i].step = 2;
      clist[i].fd = fd;

      int xb = __t_read(fd, buff, SEND_PCK_SIZE);
      if ( xb > 0 )
        clist[i].client.write((const uint8_t*)buff, xb);
      else
        clist[i].step = 99;

      clist[i].sum = xb;

      return;
    }

    String HtmlHeader = F("HTTP/1.1 404 Not Found\r\nServer: esp8266\r\nConnection: close\r\n");
    HtmlHeader += F("Access-Control-Allow-Origin: *\r\n");
    HtmlHeader += F("Content-Length: 0\r\nContent-Type: text/html\r\n\r\n");
#ifdef SERIAL_DEBUG
    Serial.printf("**** unknown request '%s'\n", RequestString.c_str());
#endif

    clist[i].client.print(HtmlHeader);
    clist[i].step = 99;
    return;
  }

  if ( clist[i].step == 2 )
  {
    char buff[SEND_PCK_SIZE];

    /* maybe some new data ? */
    while ( clist[i].client.available() )
    {
      clist[i].client.readStringUntil('\r');
      delay(10);
    }

    int xb = __t_read(clist[i].fd, buff, SEND_PCK_SIZE);
    if ( xb > 0 )
    {
      clist[i].sum += xb;
      clist[i].client.write((const uint8_t*)buff, xb);
#ifdef SERIAL_DEBUG
    Serial.print(F("["));
    Serial.print(i);
    Serial.print(F("] : "));
    Serial.print(xb);
    Serial.println("[ bytes written");
#endif
    }
    else
    {
      clist[i].step = 99;
    }

    return;
  }

  if ( clist[i].step == 3 ) /* setup.html */
  {
    int  len=clist[i].len;
    int  off=clist[i].s_done;
    int  sz = len-off;
    uint8_t  *buff=(uint8_t*)clist[i].data+off;
    if ( sz > SEND_PCK_SIZE )
      sz=SEND_PCK_SIZE;
    clist[i].client.write((const uint8_t*)buff, sz);
#ifdef SERIAL_DEBUG
    Serial.print(F("["));
    Serial.print(i);
    Serial.print(F("] : "));
    Serial.print(sz);
    Serial.println("[ bytes written");
#endif
    clist[i].s_done += sz;
    if ( clist[i].s_done == len )
    {
      clist[i].step=99;
      free(clist[i].data);
      clist[i].data=0;
    }
  }
}

static void runWiFiConfLoop( void )
{
  int       i;
  int       clnum = 0;
  WiFiClient client;

  client = server.available();

  if ( client )
  {
    for ( i = 0; i < 10; i++)
    {
      if ( !clist[i].used )
        break;
    }
    if ( i == 10 )
    {
      client.stop();
#ifdef SERIAL_DEBUG
      Serial.println("**** too many clients ****");
#endif
      return;
    }
    clist[i].client = client;
    clist[i].fd = -1;
    clist[i].step = 0; /* 0: read from client */
    clist[i].used = 1;
    clist[i].wrong_cnt = 0;
  }
  clnum = 0;
  for ( i = 0; i < 10; i++ )
  {
    if ( !clist[i].used )
      continue;
    clnum++;
    if ( clist[i].step < 90 )
      DoQueuedWiFiClient( i );
    if ( clist[i].step == 99 ) /* remove from queue */
    {
      if ( clist[i].data )
        free(clist[i].data);
      clist[i].data=0;
      if ( clist[i].fd != -1 )
        __t_close(clist[i].fd);
      clist[i].step = 98;
      delay(100);
    }
    else if ( clist[i].step == 98 ) /* remove from queue */
    {
#ifdef SERIAL_DEBUG
      Serial.println("client stop..");
#endif
      if ( clist[i].client )
        clist[i].client.stop();
      clist[i].used = 0;
    }
  }
  if ( !clnum && setup_done )  /* no open connection & setup_done */
  {
    static int rdycnt = 0;
    if ( rdycnt++ > 100 )
    {
#ifdef SERIAL_DEBUG
      Serial.println("save to EEPROM");
#endif
      Save2EEProm(0);

      writeRTC();
      ESP.restart();
      delay(100);
    }
    else
      delay(50);
  }
}

static IPAddress toIPAddress( String strg )
{
  IPAddress ip;
  int i, a, b;
  for ( i = 0, a = 0; i < 3; i++ )
  {
    b = strg.indexOf('.', a);
    ip[i] = strg.substring(a, b).toInt();
    a = b + 1;
  }
  ip[i] = strg.substring(a).toInt();

  return ip;
}

static int toPort( String val, int defport )
{
  int   pPos = -1;
  int   port = defport;

  pPos = val.indexOf(':');
  if ( pPos == -1 )
  {
    pPos = val.indexOf("%3A");
    if ( pPos != -1 )
      pPos += 3;
  }
  if ( pPos != -1 )
    port = val.substring(pPos).toInt();
  return port;
}

void prepareWiFiConf( void )
{
  int loops = 0;

  if ( config.WLan.equals("local.AP") )
  {
     WiFi.enableSTA(false);
     WiFi.mode(WIFI_AP);
     WiFi.softAP( config.DevName.c_str(), config.Password.c_str() );
     config.as_ap = 1;
     return;
  }
  WiFi.enableSTA(false);
  WiFi.enableAP(false);

  if ( !config.is_dhcp )
    WiFi.config(toIPAddress(config.IP), toIPAddress(config.Gateway), toIPAddress(config.Mask), toIPAddress(config.DNS) );
  WiFi.hostname(config.DevName);

  WiFi.begin(config.WLan.c_str(), config.Password.c_str());
}

void WiFiConnectLoop( void )
{
  if ( config.as_ap )
  {
    w_connected = 1;
    return;
  }
  w_connloops++;
  if ( w_connected )
  {
    if ( w_connloops >= 2000 )
    {
      w_connloops = 0;

      if (WiFi.status() != WL_CONNECTED)
      {
        w_connected = 0;
        iobDropLines();

        w_connloops = 1999;
      }
    }
    return;
  }
  if ( w_connloops >= 2000 )
  {
    w_connloops = 0;

    w_trys++;
    if ( w_trys == 3 )
    {
      writeRTC();
			memorizeSave();
      ESP.reset();
      return;
    }
    WiFi.disconnect();
    WiFi.hostname(config.DevName);
    WiFi.begin(config.WLan.c_str(), config.Password.c_str());
    return;
  }
  if (WiFi.status() == WL_CONNECTED)
  {
    w_connected = 1;
#ifdef SERIAL_DEBUG
    Serial.println(F(""));
    Serial.print(F("WiFi-connection to "));
    Serial.print(config.WLan);
    Serial.println(F(" : established"));

    Serial.print(F("IP Address: "));
    Serial.println(WiFi.localIP());
#endif
  }
}

void ReconfMode( void )
{
#ifdef SERIAL_DEBUG
  Serial.println("ReconfMode");
#endif
  EEPROM.begin(512);
  EEPROM.write(EE_CONFIG, CONF_RECONF);
  EEPROM.end();

  delay(5000);
  writeRTC();
	memorizeSave();
  ESP.reset();
  delay(100);
}

void _doUpdate( const char *Name )
{
#ifdef SERIAL_DEBUG
  Serial.print(F("Update="));
  Serial.println(Name);
#endif
  writeRTC();
  t_httpUpdate_return ret = ESPhttpUpdate.update(Name);
#ifdef SERIAL_DEBUG
  Serial.print(F("ret-codes: FAILED="));
  Serial.print(HTTP_UPDATE_FAILED);
  Serial.print(F(", NO_UPDATES="));
  Serial.print(HTTP_UPDATE_NO_UPDATES);
  Serial.print(F(", OK="));
  Serial.println(HTTP_UPDATE_OK);
  Serial.print(F("result="));
  Serial.print(ret);
#endif
  writeRTC();
	memorizeSave();
  ESP.reset();
  delay(100);
}
