#include <OneWire.h>

#define DS_PIN  4	/* GPIO4 */

OneWire  ds(DS_PIN);
static int save_ds_eeprom=1;
static uint32_t ds_started=0;

byte ds_addr[8];
static byte ds_scan=1;

void startDs18b20( void )
{
  float temp;
  int   t100;
  int   i;
#if 0
//  byte addr[8] = { 0x28, 0xff, 0x68, 0xde, 0xb5, 0x16, 0x03, 0x5e };
//  byte addr[8] = { 0x28, 0xff, 0x95, 0xd1, 0xb5, 0x16, 0x03, 0x8e };
  byte addr[8] = { 0x28, 0xff, 0xc9, 0x8f, 0xb3, 0x16, 0x03, 0x26 };
#endif

  if ( ds_scan )
  {
    ds.search(ds_addr);
    if (OneWire::crc8(ds_addr, 7) != ds_addr[7])
    {
       Serial.println("CRC is not valid!");
       return;
    }
    ds_scan=0;
#ifdef SERIAL_DEBUG
    Serial.print("ds-search : ");
    for(i=0;i<8;i++)
       Serial.printf(" %02x ",ds_addr[i]);
    Serial.println("");
#endif
  }
  ds.reset();

  if ( save_ds_eeprom )
  {   /* save resolution of 9bit into eeprom of ds18b20 */
    save_ds_eeprom=0;

    ds.select(ds_addr);
    ds.write(0x4E);  // set config
    ds.write(0x00);  // unused
    ds.write(0x00);  // unused
    ds.write(0x7F);  // 1f=9bit, 3f=10bit, 5f=11bit, 7f=12bit
                     // 93.75ms   187.5ms    375ms      750ms
    ds.reset();
    ds.select(ds_addr);
    ds.write(0x48);  // save into eeprom
    ds.reset();
    delay(15);
  }

  ds.select(ds_addr);
  ds.write(0x44, 1);

  ds_started=millis() | 0x80000000;
}

int16_t getTempDs18b20( void )
{
  uint32_t  now=millis() | 0x80000000;
  int       i;
  int16_t   t10=0;
#if 0
//  byte    addr[8] = { 0x28, 0xff, 0x68, 0xde, 0xb5, 0x16, 0x03, 0x5e };
//  byte      addr[8] = { 0x28, 0xff, 0x95, 0xd1, 0xb5, 0x16, 0x03, 0x8e };
  byte      addr[8] = { 0x28, 0xff, 0xc9, 0x8f, 0xb3, 0x16, 0x03, 0x26 };
#endif
  byte      data[12];
  byte      crc;

  if ( now < ds_started )
  {
     ds_started &= 0x7fffffff;    /* remove bit31 */
     ds_started |= 1;             /* make != 0 */
  }
  if ( now < ds_started + 755 )
	  return 0;					  /* wait */

  if ( ds_scan )
    return 0;

  ds.reset();
  ds.select(ds_addr);
  ds.write(0xBE);

  for ( i = 0; i < 9; i++)
  {           // we need 9 bytes
     data[i] = ds.read();

//Serial.print(data[i], HEX);
//Serial.print(" ");
  }
  crc=OneWire::crc8(data, 8);

//Serial.print(" CRC=");
//Serial.print(crc, HEX);
//Serial.println();

  if ( crc != data[8] )
    return 0;

  int16_t raw = (data[1] << 8) | data[0];
  byte cfg = (data[4] & 0x60);

//Serial.print(" CFG=");
//Serial.println(cfg,HEX);

  // at lower res, the low bits are undefined, so let's zero them
  if (cfg == 0x00) raw = raw & ~7;  // 9 bit resolution, 93.75 ms
  else if (cfg == 0x20) raw = raw & ~3; // 10 bit res, 187.5 ms
  else if (cfg == 0x40) raw = raw & ~1; // 11 bit res, 375 ms
//// default is 12 bit resolution, 750 ms conversion time
  t10=(int)((float)raw / 1.6);
#if 0
Serial.print(F("Temp: "));
Serial.print(t10/10);
Serial.print(F("."));
Serial.println(t10%10);
#endif
  return t10;
}

void ds18b20_Loop( void )
{
  if ( ds_started )
  {
    uint32_t  now=millis() | 0x80000000;
    uint32_t  diff;

    if ( now < ds_started )
    {
      ds_started &= 0x7fffffff;    /* remove bit31 */
      ds_started |= 1;             /* make != 0 */
    }
    diff=now-ds_started;

    if ( diff > 950 )
    {
      uint16_t temp=getTempDs18b20();
      double   dval;
#ifdef SERIAL_DEBUG
      String tstrg;
      String t2;

      tstrg=(temp/10);
      tstrg+=".";
      tstrg+=(temp%10);
      tstrg+="°C";

      Serial.printf("Temp : %s\n",tstrg.c_str());
#endif
      delay(10);
      dval=(double)temp/10;
      _iobSetFloatVar( IDX_TEMP, dval, SRC_BTN );

      startDs18b20();
    }
  }
}
