#include <Timezone.h>
#include <EEPROM.h>

/*
1000W :                  vorh./Tag
  1min :    1000          1440
  1h   :   60000            24
  24h  : 1440000             1

delta-use:

  0:00      :  2Bytes       (ushort)  <bis  0:59>
  1:00      :  2Bytes       (ushort)  <bis  1:59>
     ....
 23:00      :  2Bytes       (ushort)  <bis 23:59>

= 24*2 = 52 Bytes  -> EEPROM   (1Mio-Zyklen)

----
Flash:
 stamp              : 4      : timestamp-written   :   4
 9 Tage (like eepr) : 9*48   : Werte wie in eeprom : 432
 52 Wochen          : 52*1   : kWh / Woche         :  52
 10 Jahre           : 10*2   : kWh / Jahr          :  20
                                        ----------------
                                                     508  ( 1Page )
spi_flash_erase_sector();  // see EspClass::eraseConfig
spi_flash_write(offset, (uint32_t*) data, size);
spi_flash_read(offset, (uint32_t*) data, size);
    
*/


static struct
{
  uint16_t valid;		/* RTC_VALID */
  uint16_t min[61];
  uint8_t  lmin;
  uint16_t hour[24];
  uint32_t wminoff;		/* offset start akt. stunde */
  uint8_t  lhour;			/* korrespond. stunde */

  uint16_t day[30];
  uint32_t month[12];
  uint8_t  lmon;

} eedata;

#define EE_MEMORIZE 512
static int ndaycalled=0;

static void nextDay( time_t stamp )
{
  int mon=month(stamp)-1;
ndaycalled++;

	if ( mon > 11 )
		mon=11;
	/* first add day[29] to mon */
	if ( eedata.lmon < 12 )
			eedata.month[ eedata.lmon ] += eedata.day[29];
  for( int i=0; i<29; i++ )
    eedata.day[i] = eedata.day[i+1];
	eedata.day[29]=0;
	if ( eedata.lmon != mon )	/* next month */
	{
		eedata.month[ mon ]=0;
		eedata.lmon = mon;
	}
}

static void SaveEEprom( void )
{
  byte *ptr=(byte*)&eedata;
  int  i;

  EEPROM.begin(1024);
  for( i=0; i< sizeof(eedata); i++, ptr++ )
  {
    EEPROM.write(EE_MEMORIZE + i, *ptr);
  }
  EEPROM.commit();
  EEPROM.end();
}

#define MEMRTC_BASE            96
#define MEMRTC_VALID           0xb3

void memorizeSave( void )
{
  byte *ptr=(byte*)&eedata;
  int  i;
  eedata.valid = MEMRTC_VALID;

  while(system_rtc_mem_write(MEMRTC_BASE,ptr,sizeof(eedata))!=true)
        delay(1);
  yield();
  delay(5);
}

void memorizeRead( void )
{
  byte *ptr=(byte*)&eedata;
  while(system_rtc_mem_read(MEMRTC_BASE,ptr,sizeof(eedata))!=true)
        delay(1);
}

void initMemorize( void )
{
  memorizeRead();
  if ( eedata.valid != MEMRTC_VALID )
  {
	  memset(&eedata,0,sizeof(eedata));
	  eedata.lmon=99;	/* not initialized */
	  eedata.lmin=99;
	}
	else
    acs712_resetWMin( eedata.min[60] );
}

void memorizeReset( void )
{
  for( int i=0; i<24; i++ )
    eedata.hour[i]=0;
  memorizeSave();
//  SaveEEprom();
}

void	memAddValue( int wmin, time_t stamp )
{
  /* one value every hour */
  int ho=hour(stamp);
  int mi=minute(stamp);

  if ( eedata.lmin != mi )	/* next minute */
	{
		for ( int i=0; i<60; i++ )
			eedata.min[i] = eedata.min[i+1];
		eedata.lmin = mi;
	}
  eedata.min[60] = wmin;	/* raw value */

	wmin /= 200;		/* do not exceed 60000 */
									/* kWh calc: val / 300 */

	if ( wmin < eedata.wminoff )	/* there was an reset */
		eedata.wminoff = 0;

  eedata.day[29] = wmin;
  if ( eedata.lmon == 99 )
    eedata.lmon = month(stamp)-1;

  if ( ho != eedata.lhour )
  { /* nicht mehr die selbe stunde */

		uint16_t	nval= wmin - eedata.wminoff;		/* delta */
		int       oho=(int)ho-1;

    if ( oho < 0 )
      oho=23;

		if ( nval > eedata.hour[ oho ] )
			eedata.hour[ oho ] = nval;

    if ( !ho || ( ho < eedata.lhour ))
    {
      nextDay( stamp );
      acs712_resetWMin( 0 );
      wmin=0;
    }

		eedata.hour[ ho ] = 0;
		eedata.wminoff = wmin;
		eedata.lhour = ho;

//		SaveEEprom();		/* max: 10k stunden == 1.1 Jahre */
		memorizeSave();

		return;
	}
  /* delta eintragen */
  eedata.hour[ ho ] = wmin - eedata.wminoff;
}

String Mem24Values( void )
{
	String result="";
  int i;
	result += (String)(eedata.lhour);
	result += ", [";
	for( i=0; i<24; i++ )
	{
		result += (String)(eedata.hour[i]);
		if ( i<23 )
			result += ",";
	}
	result += "]";
	return result;
}

String MemDayValues( void )
{
	String result="";
  int i;
	result += "[";
	for( i=0; i<30; i++ )
	{
		result += (String)(eedata.day[i]);
		if ( i<29 )
			result += ",";
	}
	result += "]";
	return result;
}

String MemMinuteValues( void )
{
  String result="";
  int i;
  result += "[";
  for( i=0; i<60; i++ )
  {
		if ( eedata.min[i+1] < eedata.min[i] )
			result += (String)(eedata.min[i+1]);
		else
			result += (String)(eedata.min[i+1]-eedata.min[i]);
    if ( i<59 )
      result += ",";
  }
  result += "]";
  return result;
}

String MemMonthValues( void )
{
	String result="";
  int i;
	result += (String)(eedata.lmon);
	result += ",[";
	for( i=0; i<12; i++ )
	{
    if ( i==eedata.lmon )
      result += (String)(eedata.month[i] + eedata.day[29]);
    else
		  result += (String)(eedata.month[i]);
		if ( i<11 )
			result += ",";
	}
	result += "]";
	return result;
}

String memDebug( void )
{
  String res=F("<BR><BR><table cellpadding=\"3\">");;

  int i=eedata.lhour;

  res+=F("<tr><td>");
  res+=(String)(i);
  res+=F("</td><td>");
  res+=(String)(eedata.hour[i]);
  res+=F("</td></tr>");

  res+=F("<tr><td>lhour</td><td>");
  res+=(String)(eedata.lhour);
  res+=F("</td></tr><tr><td>wmin-off</td><td>");
  res+=(String)(eedata.wminoff);
  res+=F("</td></tr><tr><td>ndaycalled</td><td>");
  res+=(String)(ndaycalled);
  res+=F("</td></tr>");
  res+=F("</table>");
  return res;
}
